##########################################################################
# Copyright 2025 Cai Jianping
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
# DPCREngine succeeds PrivacyEngine.
# Just replace "PrivacyEngine" in your Opacus code with "DPCREngine",
# you will complete your private learning with our Opacus-DPCR.
##########################################################################

import dpcrpy as dcr
from opacus.grad_sample.grad_sample_module import GradSampleModule
from opacus.optimizers import DPOptimizer
from torch import nn, optim
from torch.utils.data import DataLoader
from typing import Callable, List, Tuple, Union
from dpcrpy.framework.dpcrMech import DpcrMech

from opacus import PrivacyEngine
from opacus_dpcr.dpcrOptimizer import DPCROptimizer


class DPCREngine(PrivacyEngine):

    def __init__(self, *, accountant: str = "rdp", secure_mode: bool = False,
                 dpcrGenerator: Callable[[int], DpcrMech] = lambda x: dcr.SimpleMech(1)):
        PrivacyEngine.__init__(self, accountant=accountant, secure_mode=secure_mode)
        self.dpcrGenerator = dpcrGenerator;

    def make_private(
            self,
            *,
            module: nn.Module,
            optimizer: optim.Optimizer,
            data_loader: DataLoader,
            noise_multiplier: float,
            max_grad_norm: Union[float, List[float]],
            batch_first: bool = True,
            loss_reduction: str = "mean",
            poisson_sampling: bool = True,
            clipping: str = "flat",
            noise_generator=None,
    ) -> Tuple[GradSampleModule, DPOptimizer, DataLoader]:
        # if optimizer.state_dict()['param_groups'][0]['momentum'] > 0:
        #     mom = optimizer.state_dict()['param_groups'][0]['momentum'];
        #     raise RuntimeError(f"momentum = {mom} > 0. The gradient descent with momentum does not support currently.")
        module, optimizer, data_loader = PrivacyEngine.make_private(self, module=module, optimizer=optimizer,
                                                                    data_loader=data_loader,
                                                                    noise_multiplier=noise_multiplier,
                                                                    max_grad_norm=max_grad_norm,
                                                                    batch_first=batch_first,
                                                                    loss_reduction=loss_reduction,
                                                                    poisson_sampling=poisson_sampling,
                                                                    clipping=clipping,
                                                                    noise_generator=noise_generator)
        #! PrivacyEngine.make_private 使用了错误的sample_rate，见privacy_engine.py的372-373行以及391-393行
        optimizer.attach_step_hook(
            self.accountant.get_optimizer_hook_fn(sample_rate=data_loader.sample_rate)
        )
        ###########################
        optimizer = DPCROptimizer(optimizer, self.dpcrGenerator)
        return module, optimizer, data_loader

    def __str__(self):
        res='采用了DPCREngine实现DP-SGD';
        return res;


##########################################################################
# Copyright 2025 Cai Jianping
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
# Optimizer for DPCR learning, generated through DPCREngine.
##########################################################################
from opacus.optimizers import DPOptimizer
from torch.optim import Optimizer
from typing import Callable, Optional
import torch
import dpcrpy as dcr
import numpy as np
from opacus_dpcr.tensorNoiMech import TensorGaussNoiMech
from dpcrpy.framework.dpcrMech import DpcrMech


class DPCROptimizer(Optimizer):
    def __init__(self, optimizer: DPOptimizer, dcrMechGen: Callable[[int], DpcrMech]):
        self.optimizer = optimizer
        self.dpcrs = []
        #####################################
        rou = 1.0
        accumulated_iterations = 1
        if self.optimizer.loss_reduction == "mean":
            rou /= self.optimizer.expected_batch_size * accumulated_iterations
        self.gSigma = self.optimizer.noise_multiplier * self.optimizer.max_grad_norm * rou
        lr = self.optimizer.state_dict()['param_groups'][0]['lr']
        self.gSigma *= lr
        #####################################
        self._shareId = 0;
        self.noiMeches = []
        genBlk = lambda x: dcrMechGen(x).setNoiMech(self.noiMeches[self._shareId])
        for param in self.optimizer.params:
            sz = np.array(param.size());
            noiMech = TensorGaussNoiMech(sigma0=self.gSigma, tsSize=sz, device=param.device)
            self.noiMeches.append(noiMech)
            dpcr = dcr.dpCrFw(genBlk)
            dpcr.initRelease(param)
            self.dpcrs.append(dpcr)
            self._shareId += 1;
        self.momentum_buffer_list=[None for i in range(len(self.optimizer.params))]

    def pre_step(self, closure: Optional[Callable[[], float]] = None) -> Optional[float]:
        """
        Perform actions specific to ``DPOptimizer`` before calling
        underlying  ``optimizer.step()``

        Args:
            closure: A closure that reevaluates the model and
                returns the loss. Optional for most optimizers.
        """
        if self.optimizer.params[0].grad_sample is not None:
            self.optimizer.clip_and_accumulate()
            if self.optimizer._check_skip_next_step():
                self.optimizer._is_last_step_skipped = True
                return False
            for p in self.optimizer.params:
                p.grad = p.summed_grad.view_as(p.grad)
            self.optimizer.scale_grad()
            if self.optimizer.step_hook:
                self.optimizer.step_hook(self.optimizer)
        else:
            self.accumulated_iterations=1
            self.noise_multiplier=self.optimizer.noise_multiplier
            if self.optimizer.step_hook:
                self.optimizer.step_hook(self)
        self.optimizer._is_last_step_skipped = False
        return True

    def step(self, closure: Optional[Callable[[], float]] = None) -> Optional[float]:
        if closure is not None:
            with torch.enable_grad():
                closure()
        if self.pre_step():
            lr = self.optimizer.state_dict()['param_groups'][0]['lr']
            momentum=self.optimizer.state_dict()['param_groups'][0]['momentum']
            with torch.no_grad():
                self._shareId = 0;
                for i,param in enumerate(self.optimizer.params):
                    dtParam, noiX, mse = self.dpcrs[self._shareId].dpRelease(param.grad * (-lr))

                    if momentum != 0:
                        buf = self.momentum_buffer_list[i]
                        if buf is None:
                            buf = torch.clone(noiX).detach()
                            self.momentum_buffer_list[i] = buf
                        else:
                            buf.mul_(momentum).add_(noiX, alpha=1)
                        noiX=buf

                    # if incMode:
                    param.add_(noiX)
                    # else:
                    #     param.set_(dtParam + 0)
                    self._shareId += 1;
        return self.optimizer.params;

    def zero_grad(self, set_to_none: bool = False):
        self.optimizer.zero_grad(set_to_none)

    def __repr__(self):
        return self.optimizer.__repr__()

    def state_dict(self):
        return self.optimizer.state_dict()

    def load_state_dict(self, state_dict) -> None:
        self.optimizer.load_state_dict(state_dict)

    @property
    def params(self):
        return self.optimizer.params



##########################################################################
# Copyright 2025 Cai Jianping
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
# It succeeds GaussNoiMech for supporting tensor.
##########################################################################
from dpcrpy.framework.noiMech import GaussNoiMech
import torch

class TensorGaussNoiMech(GaussNoiMech):
    def __init__(self, epsilon=None, delta=0, sens=1, sigma0=None, tsSize=(1,), device=None):
        GaussNoiMech.__init__(self, epsilon=epsilon, delta=delta, sens=sens, sigma0=sigma0)
        self.tsSize = tsSize
        self.device = device

    def genNoise(self, shape=None):
        if shape is None:
            rd = torch.normal(
                mean=0,
                std=self.sigma,
                size=list(self.tsSize),
                device=self.device
            )
            return -rd;
        rd = torch.normal(
            mean=0,
            std=self.sigma,
            size=list(shape) + list(self.tsSize),
            device=self.device
        )
        return -rd;



#!/usr/bin/env python
# -*- coding:utf-8 -*-


#############################################
# File Name: setup.py
# Author: Jianping Cai
# Mail: jpingcai@163.com
# Created Time:  2025-10-26 16:12:48
#############################################


from setuptools import setup, find_packages

setup(
    name="opacus_dpcr",
    version="1.0.0",
    keywords=["opacus", "differential-privacy", "optimizer"],
    description="Opacus extensions with DP continual release optimizers.",
    long_description="Engine and optimizer implementations enabling DP continual release workflows in Opacus.",
    license="Apache License 2.0",
    author="Jianping Cai",
    author_email="jpingcai@163.com",
    packages=find_packages(),
    include_package_data=True,
    platforms="any",
    install_requires=["numpy", "dpcrpy>=1.0.0", "opacus", "torch"],
    entry_points={}
)
